import os
import logging
import argparse
import requests
import eons as e
from zipfile import ZipFile
from distutils.dir_util import mkpath
from abc import abstractmethod
from subprocess import Popen
from subprocess import PIPE
from subprocess import STDOUT

######## START CONTENT ########
# All builder errors
class BuildError(Exception): pass


# Exception used for miscillaneous build errors.
class OtherBuildError(BuildError): pass


# Project types can be things like "lib" for library, "bin" for binary, etc. Generally, they are any string that evaluates to a different means of building code.
class ProjectTypeNotSupported(BuildError): pass


class EBBS(e.Executor):

    def __init__(self):
        super().__init__(name="eons Basic Build System", descriptionStr="A hackable build system for all languages!")

        self.RegisterDirectory("language")
        self.RegisterDirectory("inc/language")
        self.RegisterDirectory("ebbs/inc/language")

    #Override of eons.Executor method. See that class for details
    def RegisterAllClasses(self):
        super().RegisterAllClasses()
        self.RegisterAllClassesInDirectory(os.path.join(os.path.dirname(os.path.abspath(__file__)), "language"))

    #Override of eons.Executor method. See that class for details
    def AddArgs(self):
        super().AddArgs()
        self.argparser.add_argument('dir', type = str, metavar = '/project/build', help = 'path to build folder', default = '.')
        self.argparser.add_argument('-l','--language', type = str, metavar = 'cpp', help = 'language of files to build', dest = 'lang')
        self.argparser.add_argument('--repo-store', type=str, default='./ebbs/inc/language', help='file path for storing downloaded packages', dest = 'store')
        self.argparser.add_argument('--repo-url', type = str, default='https://api.infrastructure.tech/v1/package', help = 'package repository for additional languages', dest = 'url')
        self.argparser.add_argument('--repo-username', type = str, help = 'username for http basic auth', dest = 'username')
        self.argparser.add_argument('--repo-password', type = str, help = 'password for http basic auth', dest = 'password')

    #Override of eons.Executor method. See that class for details
    def ParseArgs(self):
        super().ParseArgs()

        if (not self.args.lang):
            self.ExitDueToErr("You must specify a language.")

    #Override of eons.Executor method. See that class for details
    def UserFunction(self, **kwargs):
        super().UserFunction(**kwargs)
        self.Build()

    #Build things!
    def Build(self):
        try:
            builder = self.GetRegistered(self.args.lang)
        except:
            logging.debug(f'Builder for {self.args.lang} not found. Trying to download from repository.')
            try: #again
                self.DownloadPackage(f'build_{self.args.lang}')
                builder = self.GetRegistered(self.args.lang)
            except Exception as e:
                logging.error(f'Could not find builder for {self.args.lang}: {e}')
                raise OtherBuildError(f'Could not get builder for {self.args.lang}')
                return #just for extra safety.

        repoData = {}
        if (self.args.store and self.args.url and self.args.username and self.args.password):
            repoData = {
                'store': self.args.store,
                'url': self.args.url,
                'username': self.args.username,
                'password': self.args.password
            }

        builder(dir = self.args.dir, repo = repoData, **self.extraArgs)

    #Attempts to download the given package from the repo url specified in calling args.
    #Will refresh registered builders upon success
    #RETURNS void
    #Does not guarantee new builders are made available; errors need to be handled by the caller.
    def DownloadPackage(self, packageName):

        url = f'{self.args.url}/download?package_name={packageName}'

        auth = None
        if self.args.username and self.args.password:
            auth = requests.auth.HTTPBasicAuth(self.args.username, self.args.password)

        packageQuery = requests.get(url, auth=auth)

        if (packageQuery.status_code != 200 or not len(packageQuery.content)):
            logging.error(f'Unable to download {packageName}')
            #TODO: raise error?
            return #let caller decide what to do next.

        if (not os.path.exists(self.args.store)):
            logging.debug(f'Creating directory {self.args.store}')
            mkpath(self.args.store)

        packageZip = os.path.join(self.args.store, f'{packageName}.zip')

        logging.debug(f'Writing {packageZip}')
        open(packageZip, 'wb+').write(packageQuery.content) #TODO: close?
        if (not os.path.exists(packageZip)):
            logging.error(f'Failed to create {packageZip}')
            return

        logging.debug(f'Extracting {packageZip}')
        ZipFile(packageZip, 'r').extractall(f'{self.args.store}') #TODO: close?

        logging.debug(f'Registering classes in {self.args.store}')
        self.RegisterAllClassesInDirectory(self.args.store)

class Builder(e.UserFunctor):
    def __init__(self, name=e.INVALID_NAME()):
        super().__init__(name)
        
        self.requiredKWArgs.append("dir")

        self.supportedProjectTypes = []

        #TODO: project is looking an awful lot like a Datum.. Would making it one add functionality?
        self.projectType = "bin"
        self.projectName = e.INVALID_NAME()

    #Build things!
    #Override this or die.
    @abstractmethod
    def Build(self):
        raise NotImplementedError

    #Projects should have a name of {project-type}_{project-name}.
    #For information on how projects should be labelled see: https://eons.dev/convention/naming/
    #For information on how projects should be organized, see: https://eons.dev/convention/uri-names/
    def PopulateProjectDetails(self):
        details = os.path.basename(os.path.abspath(os.path.join(self.buildPath,"../"))).split("_")
        self.projectType = details[0]
        if (len(details) > 1):
            self.projectName = '_'.join(details[1:])
        
    #Sets the build path that should be used by children of *this.
    #Also sets src, inc, lib, and dep paths, if they are present.
    def SetBuildPath(self, path):
        self.buildPath = path

        #TODO: Consolidate this code with more attribute hacks?
        rootPath = os.path.abspath(os.path.join(self.buildPath, "../"))
        if (os.path.isdir(rootPath)):
            self.rootPath = rootPath
        else:
            self.rootPath = None
        srcPath = os.path.abspath(os.path.join(self.buildPath, "../src"))
        if (os.path.isdir(srcPath)):
            self.srcPath = srcPath
        else:
            self.srcPath = None
        incPath = os.path.abspath(os.path.join(self.buildPath, "../inc"))
        if (os.path.isdir(incPath)):
            self.incPath = incPath
        else:
            self.incPath = None
        depPath = os.path.abspath(os.path.join(self.buildPath, "../dep"))
        if (os.path.isdir(depPath)):
            self.depPath = depPath
        else:
            self.depPath = None
        libPath = os.path.abspath(os.path.join(self.buildPath, "../lib"))
        if (os.path.isdir(srcPath)):
            self.libPath = libPath
        else:
            self.libPath = None

    #Hook for any pre-build configuration
    def PreBuild(self, **kwargs):
        pass

    #Hook for any post-build configuration
    def PostBuild(self, **kwargs):
        # TODO: Do we need to clear self.buildPath here?
        pass

    def UserFunction(self, **kwargs):
        self.SetBuildPath(kwargs.get("dir"))
        self.PopulateProjectDetails()
        self.PreBuild(**kwargs)
        if (len(self.supportedProjectTypes) and self.projectType not in self.supportedProjectTypes):
            raise ProjectTypeNotSupported(f"{self.projectType} is not supported. Supported project types for {self.name} are {self.supportedProjectTypes}")
        logging.info(f"Using {self.name} to build {self.projectName}, a {self.projectType}")
        self.Build()
        self.PostBuild(**kwargs)

    #RETURNS: an opened file object for writing.
    #Creates the path if it does not exist.
    def CreateFile(self, file, mode="w+"):
        mkpath(os.path.dirname(os.path.abspath(file)))
        return open(file, mode)

    #Run whatever.
    #DANGEROUS!!!!!
    #TODO: check return value and raise exceptions?
    #per https://stackoverflow.com/questions/803265/getting-realtime-output-using-subprocess
    def RunCommand(self, command):
        p = Popen(command, stdout = PIPE, stderr = STDOUT, shell = True)
        while True:
          line = p.stdout.readline()
          if (not line):
            break
          print(line.decode('ascii')[:-1]) #[:-1] to strip excessive new lines.




import os
import logging
import shutil
import requests
from ebbs import Builder
from ebbs import OtherBuildError

class publish(Builder):
    def __init__(self, name="Wordpress Plugin Builder"):
        super().__init__(name)

        self.requiredKWArgs.append("repo")
        self.requiredKWArgs.append("--version")
        # self.requiredKWArgs.append("--visibility") # this is optional. Defaults to private

        self.supportedProjectTypes = [] #all

    def PreBuild(self, **kwargs):
        self.repo = kwargs.get("repo")
        if (not len(self.repo)):
            raise OtherBuildError(f'Repo credentials required to publish package')

        self.targetFileName = f'{self.projectName}.zip'
        self.targetFile = os.path.join(self.rootPath, self.targetFileName)

        self.packageName = f'{self.projectType}_{self.projectName}'

        self.packageVisibility = 'private'
        if ("--visibility" in kwargs):
            self.packageVisibility = kwargs.get("--visibility")

        self.packageType = ''
        if ("--package-type" in kwargs):
            self.packageType = kwargs.get("--package-type")

        self.desciption = ''
        if ("--desciption" in kwargs):
            self.desciption = kwargs.get("--desciption")

        self.requestData = {
            'package_name': self.packageName,
            'version': kwargs.get("--version"),
            'visibility': self.packageVisibility
        }
        if (self.packageType):
                self.requestData['package_type'] = self.packageType
        if (self.desciption):
            self.requestData['description'] = self.desciption

    # Required Builder method. See that class for details.
    def Build(self):
        logging.debug("Creating archive")
        if (os.path.exists(self.targetFile)):
            os.remove(self.targetFile)

        shutil.make_archive(self.targetFile[:-4], 'zip', self.buildPath)
        logging.debug("Achive created")

        logging.debug("Uploading archive to repository")
        files = {
            'package': open(self.targetFile, 'rb')
        }
        logging.debug(f'Request data: {self.requestData}')
        packageQuery = requests.post(f"{self.repo['url']}/publish", auth=requests.auth.HTTPBasicAuth(self.repo['username'], self.repo['password']), data=self.requestData, files=files)

        if (packageQuery.status_code != 200):
            logging.error(f'Failed to publish {self.projectName}')
            raise OtherBuildError(f'Failed to publish {self.projectName}')

        logging.info(f'Successfully published {self.projectName}')


import os
import logging
from ebbs import Builder

#Class name is what is used at cli, so we defy convention here in favor of ease-of-use.
class cpp(Builder):
    def __init__(self, name="C++ Builder"):
        super().__init__(name)
        
        self.supportedProjectTypes.append("lib")
        self.supportedProjectTypes.append("bin")
        self.supportedProjectTypes.append("test")

        self.valid_cxx_extensions = [
            ".cpp",
            ".h"
        ]
        self.valid_lib_extensions = [
            ".a",
            ".so"
        ]

    #Required Builder method. See that class for details.
    def Build(self):
        os.chdir(self.buildPath)
        self.GenCMake()
        self.CMake()
        self.Make()

    def get_cxx_files (self, directory, seperator=" "):
        ret = ""
        for root, dirs, files in os.walk(directory):
            for f in files:
                name, ext = os.path.splitext(f)
                if (ext in self.valid_cxx_extensions):
                    # logging.info(f"    {os.path.join(root, f)}")
                    ret += f"{os.path.join(root, f)}{seperator}"
        return ret[:-1]
    
    def get_libs(self, directory, seperator=" "):
        ret = ""
        for file in os.listdir(directory):
            if not os.path.isfile(os.path.join(directory, file)):
                continue
            name, ext = os.path.splitext(file)
            if (ext in self.valid_lib_extensions):
                ret += (f"{name[3:]}{seperator}")
        return ret[:-1]

    def GenCMake(self):
        #Write our cmake file
        cmake_open = '''
cmake_minimum_required (VERSION 3.1.1)
set (CMAKE_CXX_STANDARD 11)
include_directories(${CMAKE_SOURCE_DIR}/../inc)
'''

        cmake_file = open("CMakeLists.txt", "w")
        cmake_file.write(f"{cmake_open}\n")
        cmake_file.write(f"project ({self.projectName})\n")

        if (self.projectType in ["bin", "test"]):
            logging.info("Addind binary specific code")
            
            cmake_file.write (f"add_executable ({self.projectName} {self.get_cxx_files ('../src')})\n")
            cmake_file.write ('''
include_directories(${CMAKE_SOURCE_DIR}/../lib)
set(THREADS_PREFER_PTHREAD_FLAG ON)
find_package(Threads REQUIRED)
''')
            cmake_file.write (f"target_link_directories({self.projectName} " + "PUBLIC ${CMAKE_SOURCE_DIR}/../lib)\n")
            cmake_file.write (f"target_link_libraries({self.projectName} Threads::Threads {self.get_libs('../lib')})")

        if (self.projectType in ["lib", "mod"]):
            logging.info("Addind library specific code")

            #TODO: support windows install targets
            src_path = "/usr/local/lib"
            inc_path = "/usr/local/include/{self.projectName}"

            cmake_file.write (f"add_library ({self.projectName} STATIC {self.get_cxx_files('../src')})\n")
            cmake_file.write (f"set_target_properties({self.projectName} PROPERTIES PUBLIC_HEADER \"{self.get_cxx_files('../inc',';')}\")\n")
            cmake_file.write (f"INSTALL(TARGETS {self.projectName} LIBRARY DESTINATION {src_path} PUBLIC_HEADER DESTINATION {inc_path})\n")

        cmake_close = '''
'''
        cmake_file.write (f"{cmake_close}")

        cmake_file.close()

    def CMake(self):
        self.RunCommand("cmake .")

    def Make(self):
        self.RunCommand("make")




import os
import logging
from distutils.file_util import copy_file
from distutils.dir_util import copy_tree, mkpath
from ebbs import Builder

#Class name is what is used at cli, so we defy convention here in favor of ease-of-use.
class py(Builder):
    def __init__(self, name="Python Builder"):
        super().__init__(name)
        
        self.supportedProjectTypes.append("lib")
        self.supportedProjectTypes.append("bin")

        self.validPyExtensions = [
            ".py"
        ]

    def PreCall(self, **kwargs):
        super().PreCall(**kwargs)
        self.outFile = None
        self.decomposedFiles = []
        self.imports = [] #all import statements
        self.consolidatedContents = [] #all file contents. FIXME: This doesn't scale but need to write imports first.

    #Required Builder method. See that class for details.
    def Build(self):
        self.buildPath = os.path.abspath(os.path.join(self.buildPath, self.projectName))
        mkpath(self.buildPath)
        os.chdir(self.buildPath)
        logging.info(f"Using build path {self.buildPath}")
        self.outFile = self.CreateFile(f"{self.projectName}.py")
        logging.info(f"Consolidating python files from {self.srcPath}")
        self.DecomposePyFiles(self.srcPath)
        self.WriteImports()
        self.outFile.write(f'''
######## START CONTENT ########
''')
        self.WriteContents()
        self.outFile.close()
        if (self.projectType == "lib"):
            self.MakeLib()
        elif (self.projectType == "bin"):
            self.MakeBin()
        self.CopyIncludes()
        # If we can build our prepared project, let's do it!
        if (os.path.isfile(os.path.join(self.rootPath, "setup.cfg"))):
            logging.info(f"Begining python build process")
            os.chdir(self.rootPath)
            self.BuildPackage()
            self.InstallPackage()
        logging.info("Complete!")

    #Adds an import line to *this.
    #Prevents duplicates.
    def AddImport(self, line):
        if (line in self.imports):
            return
        self.imports.append(line)

    #Decompose a python file into imports and content.
    #Both are currently stored in member variables.
    #Recursive to account for dependencies.
    #Does not operate on the same file more than once.
    def Decompose(self, pyFile):
        absPyFilePath = os.path.abspath(pyFile)
        if (absPyFilePath in self.decomposedFiles):
            logging.debug(f"Already decomposed {absPyFilePath}")
            return
        logging.debug(f"Starting to decompose {absPyFilePath}")
        py = open(pyFile, "r")
        for line in py:
            if (line.startswith("from") or line.startswith("import")): #handle import parsing
                spaced = line.split(' ')
                if (spaced[1].startswith(".")): #Decompose dependency first.
                    #TODO: Isn't "...", etc. a thing with relative imports?
                    dependency = spaced[1].replace(".", "/") + ".py"
                    if (dependency.startswith("/") and not dependency.startswith("//")):
                        dependency = dependency[1:]
                    dependency = dependency.replace("//", "../")
                    logging.debug(f"Found dependency {dependency}")
                    self.Decompose(os.path.join(os.path.dirname(pyFile), dependency))
                    continue
                multiports = line.split(",")
                if (len(multiports) > 1): #break out multiple imports for duplicate checking
                    # logging.debug(f"Breaking up multiple imports from {line}")
                    begin = " ".join(multiports[0].split(" ")[:-1])
                    # logging.debug(f"Beginning = {begin}")
                    #TODO: Do we need to support "\r\n" for windows?
                    #TODO: What's up with all this [:-1]+"\n" nonsense? Why does that invisible line ending change the uniqueness of the string (i.e. what is the line ending if not "\n")?
                    for i, imp in enumerate(multiports):
                        if (i == 0):
                            self.AddImport(imp + "\n")
                            continue
                        elif (i == len(multiports)-1):
                            self.AddImport(begin + imp[:-1] + "\n")
                        else:
                            self.AddImport(begin + imp + "\n")
                        # logging.debug(f"Got new import: {begin + imp}")
                else:
                    self.AddImport(line[:-1] + "\n")

            #TODO: Strip comments and newlines for smaller file footprint

            else: #content line
                #FIXME: See above FIXME. This should be self.outFile.write(line) but imports need to be written first.
                #FIXME: Need to enforce each line ending with a newline without things becoming weird.
                self.consolidatedContents.append(line)
        self.consolidatedContents.append("\n")
        self.decomposedFiles.append(absPyFilePath)
        logging.debug(f"Finished decomposing {absPyFilePath}")

    #Walk a directory and Decompose all python files in it.
    def DecomposePyFiles(self, directory):
        for root, dirs, files in os.walk(directory):
            for f in files:
                name, ext = os.path.splitext(f)
                if (ext in self.validPyExtensions):
                    # logging.info(f"    {os.path.join(root, f)}")
                    self.Decompose(os.path.join(root,f))

    #Dump contents of import member buffer to disk.
    def WriteImports(self):
        for imp in self.imports:
            self.outFile.write(imp)

    #Dump contents of content member buffer to disk.
    def WriteContents(self):
        for line in self.consolidatedContents:
            self.outFile.write(line)

    #Makes an empty init file
    def MakeEmptyInitFile(self, path):
        initFile = self.CreateFile(os.path.join(path, "__init__.py"))
        initFile.write(f'''
''')
        initFile.close()

    #Makes package a library
    def MakeLib(self):
        initFile = self.CreateFile(os.path.join(self.buildPath, "__init__.py"))
        initFile.write(f'''from .{self.projectName} import *
''')
        initFile.close()

    #Makes package executable
    def MakeBin(self):
        logging.info(f"Adding binary specific code.")
        initFile = self.CreateFile("__init__.py")
        #TODO: Support projects that aren't capitalized acronyms. For now, though, this is easy.
        initFile.write(f'''#!/usr/bin/env python3
from .{self.projectName} import *
{self.projectName} = {self.projectName.upper()}()
''')
        initFile.close()

        mainFile = self.CreateFile("__main__.py")
        mainFile.write(f'''from . import {self.projectName}
if __name__ == '__main__':
    {self.projectName}()
''')
        mainFile.close()

    #Copy all files from the project include directory into our build folder.
    def CopyIncludes(self):
        if (self.incPath):
            logging.info("Copying includes")
        else:
            return

        for thing in os.listdir(self.incPath):
            thingPath = os.path.join(self.incPath, thing)
            destPath = os.path.join(self.buildPath, thing)
            if os.path.isfile(thingPath):
                copy_file(thingPath, destPath)
            elif os.path.isdir(thingPath):
                copy_tree(thingPath, destPath)
        for root, dirs, files in os.walk(self.buildPath):
            for d in dirs:
                self.MakeEmptyInitFile(os.path.join(root,d))

    #Builds the thing.
    def BuildPackage(self):
        self.RunCommand("python -m build")

    #Installs the built package!
    def InstallPackage(self):
        self.RunCommand("pip install . -U")